/*
Copyright 2018 The Kubernetes Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

package plugin

import (
	"strings"

	"github.com/spf13/cobra"
	"github.com/spf13/pflag"
)

var reservedFlags = map[string]struct{}{
	"alsologtostderr":          {},
	"as":                       {},
	"as-group":                 {},
	"cache-dir":                {},
	"certificate-authority":    {},
	"client-certificate":       {},
	"client-key":               {},
	"cluster":                  {},
	"context":                  {},
	"help":                     {},
	"insecure-skip-tls-verify": {},
	"kubeconfig":               {},
	"log-backtrace-at":         {},
	"log-dir":                  {},
	"log-flush-frequency":      {},
	"logtostderr":              {},
	"match-server-version":     {},
	"n":                        {},
	"namespace":                {},
	"password":                 {},
	"request-timeout":          {},
	"s":                        {},
	"server":                   {},
	"stderrthreshold":          {},
	"token":                    {},
	"user":                     {},
	"username":                 {},
	"v":                        {},
	"vmodule":                  {},
}

// full paths of commands that should not show up in the plugin
var commandsToSkip = map[string]struct{}{
	"svcat install": {},
}

// Manifest is the root structure of the kubectl plugin manifest.
type Manifest struct {
	Plugin `json:",inline"`
}

// Plugin describes a command exposed by the plugin.
type Plugin struct {
	// Name of the command for the help text. Required.
	Name string `json:"name"`

	// Use is the one-line description of how the command is used.
	Use string `json:"use"`

	// ShortDesc is a one-line description of the command. Required.
	ShortDesc string `json:"shortDesc"`

	// LongDesc is the optional full description of the command.
	LongDesc string `json:"longDesc,omitempty"`

	// Example contains optional examples of how to use the command.
	Example string `json:"example,omitempty"`

	// Command that the kubectl plugin runner should execute. Required.
	Command string `json:"command"`

	// Flags supported by the command.
	Flags []Flag `json:"flags,omitempty"`

	// Tree of child commands.
	Tree []Plugin `json:"tree,omitempty"`
}

// Flag describes a flag exposed by a plugin command.
type Flag struct {
	// Name of the flag. Required.
	Name string `json:"name"`

	// Shorthand flag, must be a single character.
	Shorthand string `json:"shorthand,omitempty"`

	// Desc of the flag for the help text. Required.
	Desc string `json:"desc"`

	// DefValue is the default value to use when the flag is not specified.
	DefValue string `json:"defValue,omitempty"`
}

// Load a cli command into the plugin manifest structure.
func (m *Manifest) Load(rootCmd *cobra.Command) {
	m.Plugin = m.convertToPlugin(rootCmd)
}

func (m *Manifest) convertToPlugin(cmd *cobra.Command) Plugin {
	p := Plugin{}

	p.Name = strings.Split(cmd.Use, " ")[0]
	p.Use = cmd.Use
	p.ShortDesc = cmd.Short
	if p.ShortDesc == "" {
		p.ShortDesc = " " // The plugin won't validate if empty
	}
	p.LongDesc = cmd.Long
	p.Command = "./" + cmd.CommandPath()
	p.Example = cmd.Example

	p.Flags = []Flag{}
	cmd.Flags().VisitAll(func(flag *pflag.Flag) {
		result := m.convertToFlag(flag)
		if result != nil {
			p.Flags = append(p.Flags, *result)
		}
	})

	p.Tree = []Plugin{}
	for _, subCmd := range cmd.Commands() {
		if _, skip := commandsToSkip[subCmd.CommandPath()]; !skip {
			p.Tree = append(p.Tree, m.convertToPlugin(subCmd))
		}
	}
	return p
}

func (m *Manifest) convertToFlag(src *pflag.Flag) *Flag {
	if _, reserved := reservedFlags[src.Name]; reserved {
		return nil
	}

	dest := &Flag{
		Name: src.Name,
		Desc: src.Usage,
	}

	if _, reserved := reservedFlags[src.Shorthand]; !reserved {
		dest.Shorthand = src.Shorthand
	}

	return dest
}
